/****************************************************************************
 *                                                                          *
 * This program called cvi (C verbal interpreter) looks for information     *
 * about variables and data types in an ANSI C program.                     *
 * Copyright (c) 1999 Chepelianskii Alexeii.                                *
 *                                                                          *
 * This program is free software; you can redistribute it and/or modify     *
 * it under the terms of the GNU General Public License as Published by     *
 * the Free Software Fondation; either version 2 of the License, or (at     *
 * your option) any later version.                                          *
 *                                                                          *
 * This program is distributed in the hope it will be useful, but WITHOUT   *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    *
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License     *
 * for more details.                                                        *
 *                                                                          *
 * You should have received a copy of the GNU General Public License along  *
 * with this program; if not, write to the Free Software Foundation, Inc.,  *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                  *
 *                                                                          *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "trad.h"

/* external variables not in the .h file. */
extern char *lex;
extern int typelex;
extern list *MainL;
extern int lirelex(void);
extern void remettre(int);

/* helper routines for liretype(). */
static void ecriretype(char *str, struct Type *inf, int nelem, char *namestr);
static void parsetype(struct Type *info, int nelem, int *correct);
static int tryTypedef(void);
static void lire(Sortie *,struct Type *,int *,int *,int,char **, char *,int);
static Struc *DefStruc(char *name ,int status, int Isdef, char *Where, Entry);

/* liretype : Parses an ANSI/ISO C type defines structures, enumerations, 
   typedefs.
   Be careful modify lex and typelex. */
int liretype(int niv, struct Type *infotype, int n, struct Sortie *sor,
	                                              char *func)
{
     int cor = 1;
     int trv = 0, struc = 0;
     int TypedefFound = 0;
     char *nomstr = NULL;
     
     TypedefFound = tryTypedef();
     lire(sor, infotype, &trv, &struc, n, &nomstr, func, TypedefFound);     
     if (!TypedefFound) {
        TypedefFound = tryTypedef();
	lirelex();
     }
     if (trv && !struc)
        remettre(typelex);
     else if (!trv) {
        error("unknown data type, line %d", nlignes);
        return 0;
     }
     if (struc == DEFSTRUC)
        return DEFSTRUC;
     if (!strcmp(lex, "typedef")) {
        TypedefFound = 1;
	lirelex();
     }
     parsetype(infotype, n, &cor);
     if (!cor)
        return cor;
     ecriretype(sor->var.type, infotype, n, nomstr);
     if (infotype[Void].ntrouve && infotype[Void].ntrouve != INDULGENT)
        sor->erreurpos = 1;
     if (TypedefFound && niv > DECL) {
        printf(" declaration of a type in a parameter list\n");
	return 0;
     } else if (TypedefFound)
        return TYPEDEF;
     return cor;
}


inline static defineStr(struct Struc *pn)
{
    struct infovar info;
    char *temp = NULL;
    int res = 0, Isparsed = 0;
    Prog *pr = (Prog *) getG(MainL);

    nblocs++;
    mstrcpy(&temp, "structure/union ");
    mstrcat(&temp, pn->str.nom);
    if (lirelex() == '}')
       return 1;
    else if (typelex != NOM && typelex != '}') {
       error("syntax line %d", nlignes);
       return 0;
    } 
    for (; typelex == NOM && IsType(lex) ; lirelex()) {
       remettrech(lex);
       res = trad(&pn->obj.dec, &info, temp);
       /* A structure can contain only variable definitions. */
       if (res != DECLARATION)
	  return 0;
       Isparsed++;
       if (res == DEFSTRUC)
	  continue;
       /* to forbid structures which contain themselves in their definition */
       if (info.str != NULL && motcmp(info.str->str.nom, pn->str.nom)) {
	  if ( info.whatIs == VAR) {
	     error("reference to an incomlete type in %s", pn->str.nom);
	     return 0;
	  }
       }
    }
    nblocs--;
    newbloc(pr != NULL ? *pr->var : NULL);
    free((void *) temp);
    return Isparsed;
}

void InstallEnum(char *name, char *func)
{
    struct infovar var = {0};
    Prog *pr = (Prog *) getG(MainL);
    
    mstrcpy(&var.fonction, pr->func);
    vastrcat(2,&var.text, ENUMT, func);
    mstrcpy(&var.nom, name);
    mstrcpy(&var.strtype, "");
    var.ligne = nlignes;
    var.status = nblocs;
    *pr->var = ajoutarbre(*pr->var, &var);
}


static char *considerEnumList(char *nom)
{
    static int size = 150;
    static char *list = NULL;

    if (list == NULL)
       list = (char *) malloc(size * sizeof(char));
    *list = '\0';
    while (lirelex() != '}') {
       if (typelex == NOM) {
	  int dd;
	  if ((dd = size - strlen(lex) - strlen(list) - 3) < 0)
	     if ((list = (char *) realloc((void *) list, size -= dd)) == NULL)
	        AllocError();
	  strcat(list, lex);
	  strcat(list, " ");
	  InstallEnum(lex, nom);
       } else if (typelex != ',') {
	  printf("%c\n", typelex);
	  error("syntax ");
	  exit(1);
       }
    }
    return list;
}

/* parse the structure named lex, returns 0 if an error occur */
static int parseStr(Sortie *sor, char *func,const struct Type *infotype,
       char **nomstr, UserTypes crit)
{ 
    struct Struc *pn;
    static defaultname = 0;       /* for unnamed structures/unions/enum */
    int continuer = 0;
    int status = nblocs;
    Entry forStr;

    if (crit != DEFENUM && crit != DEFSTRUC) {
       error("parseStr : wrong argument\n");
       return 0;
    } else
       forStr = (crit == DEFSTRUC) ? STR_INF : ENUM_INF;
    if (typelex == '{') {
       /* insert the following string into the imput stack */
       char temp[50] = DEFAULT;
       remettrecar('{');
       remettrech(strcat(temp, conv(++defaultname)));
       continuer = 1;
    }
    if (typelex == NOM || continuer) {
       if (continuer)
	  lirelex();
       if ((pn = consulter(lex, forStr)) != NULL && 
	                            lirelex() != '{') {
	  /* If the structure/union/enum has been already declared */
	  if (forStr == STR_INF && ((pn->str.whatIs == UNION && 
              !infotype[Union].ntrouve || !pn->str.whatIs && 
	      !infotype[Struct].ntrouve))) {
	     error("%s isn't a%s %s", pn->str.nom, pn->str.whatIs ? "" : "n", 
		   (pn->str.whatIs) ? "structure" : "union");
	     return 0;
	  }
	  if (forStr == STR_INF)
	     sor->var.str = pn;
	  *nomstr = pn->str.nom;
	  mstrcat(&sor->var.strtype, *nomstr);
       } else {
	  mstrcat(&sor->var.nom, lex);
	  
	  if (pn == NULL)
	     lirelex();
	  if (dicho(lex, motcle, NREST)) {
	     error("%s is a keyword in C, line %d" ,lex, nlignes);
	     return 0;
	  }
	  if (typelex == ';') {
	     if ((pn = DefStruc(sor->var.nom, status, NON, func, forStr)) 
		                                                     == NULL)
	        return 0;
	     if (forStr == STR_INF && infotype[Union].ntrouve)
	        pn->str.whatIs = UNION;
	     return crit;
	  } else if (typelex != '{') {
	     if ((pn = DefStruc(sor->var.nom, status, NON, func, forStr)) 
		                                                     == NULL)
	        return 0;
	     if (forStr == STR_INF && infotype[Union].ntrouve)
	        pn->str.whatIs = UNION;
	     *nomstr = pn->str.nom;
	     mstrcat(&sor->var.strtype, *nomstr);
	     if (forStr == STR_INF)
		   sor->var.str = pn;
	  } else {
	     int Isparsed = 0;
	     
	     if ((pn = DefStruc(sor->var.nom, status, OUI, func, forStr)) 
		                                                     == NULL)
	        return 0;
	     if (forStr == STR_INF) {
	        Isparsed = defineStr(pn);
		if (infotype[Union].ntrouve)
		   pn->str.whatIs = UNION;
		if (typelex != '}' || !Isparsed)
		   return 0;
	     } else if ((pn->str.text = 
			 dupchaine(considerEnumList(pn->str.nom))) == NULL)
		   AllocError();
	     if (lirelex() == ';') {
	        if (forStr == STR_INF && continuer) {
		   error("unnamed struct/union that defines "
			 "no instances line %d", nlignes);
		   return 0;
		}
		return crit;
	     }
	     *nomstr = pn->str.nom;
	     mstrcat(&sor->var.strtype, *nomstr);
	     if (forStr == STR_INF)
		   sor->var.str = pn;
	  }
       }
    } else {
       printf(" erreur de syntaxe ligne %d\n", nlignes);
       return 0;
    }
}


/* returns 1 if a typedef is found */
static tryTypedef(void)
{
     if (typelex == NOM && !strcmp(lex, "typedef"))
        return OUI;
     remettre(typelex);
     return NON;
}

inline static void add(struct Type *, int *, int nElems);
/* read the type of a variable from the imput */
static void lire(Sortie *s, struct Type *pt, int *found, int *struc, int n
		                  , char **nst, char *fnc, int type)
{
    struct Type *p;
    Struc *def;
    int discovered = 0;

    if (lirelex() != NOM)
       return;
    while ((p = dichoType(lex, pt, n)) != NULL ||
	                           (def = consulter(lex, TP_INF)) != NULL) {
       (*found)++;
       lirelex();

       if (p == NULL && def != NULL && !*struc) {	   
	   int user = def->obj.type[User];
	   if (discovered && type) {
	      /* to allow somewhat like this : 
		 typedef int I; void f(void) { typedef float I };
		 in fact : when f::I is parsed it may be 
		 considered as a type not as a variable name.
		 hence that logical test.
	      */
	      remettrecar(typelex);
	      typelex = NOM;
	      break;
	   }
	   add(pt, def->obj.type, n);
	   vastrcat(2, &s->var.strtype, def->str.nom, " ");
	   s->var.str = def->str.str;
	   if (def->obj.type[Struct] || def->obj.type[Union] || 
	       def->obj.type[Enum])
	      *nst = def->str.type;
	   /* no type verification for a typedef containing USER */
	   if (discovered && !user) {
	      error("typedef with an other type line %d\n", nlignes);
	      remettrecar(0);
	   } else if (user && typelex == NOM)
	      continue;
	   break;
	} else if (*struc) {
	   break;
	} else {
	   int cor;

	   p->ntrouve++;
	   if (p->status == STRUCT || p->status == ENUM) {
	      UserTypes crit = (p->status == STRUCT) ? DEFSTRUC : DEFENUM;

              vastrcat(2, &s->var.strtype, p->nomtype, " ");
              if ((cor = parseStr(s, fnc, pt, nst, crit)) == crit) {
                 *struc = DEFSTRUC;
                 return;
              } else if (!cor)
		 remettrecar(0); /* cause a syntax error */
	   } else if (p->ntrouve == 1)
	      mstrcat(&s->var.strtype, p->nomtype);
	   if (p->status != ACCSMOD && p->status != STOCK)
	      discovered++;
	}
        mstrcat(&s->var.strtype, " ");
	if (typelex != NOM)
	   break;
    }
}
      
static void parsetype(struct Type *info, int n, int *cor)
{
     register tr, nu, cond;
     struct Type *mem = info;

     for (tr = nu = cond = 0 ;--n; info++)
        if (info->status == STOCK && info->ntrouve > 1) {
	   error("wrong storage class specifier %d",nlignes);
	   *cor = 0;
	} else if (info->status == TYPE && (info->ntrouve > 1 || 
					           tr && info->ntrouve)) {
	      error("conflicting types for the same variable "
		    "line %d.", nlignes);
	      *cor = 0;
	} else if (info->status == QUAL && info->ntrouve) {
	   if (mem[Void].ntrouve) {
	      error("void with another type");
	      *cor = 0;
	   }
	   /* to set the default integer type */
	   nu++;
        } else if (info->ntrouve) {
	   if (info->status == STOCK || info->status == ACCSMOD)
	      nu++;
	   else if (info->status != INDULGENT)
	   /* _1 : to know if we have already found a basic data type */
	   /* _2 : to understand that short a; and short int a; have the
	      same type */
	      tr++;
	}
     if (mem[Signed].ntrouve && mem[Unsigned].ntrouve ||
	 (cond = mem[Long].ntrouve && mem[Short].ntrouve)) {
        error("type declared as %s",
		cond ? "'long' and 'short'" : "'signed' and 'unsigned'");
        *cor =  0;
     }
     if (!tr && nu)
        mem[Int].ntrouve = 1;
}

/* write a string to describe the type defined by 'info'.
   Does not consider storage class specifiers as the following
   code is correct :
       static A;
       extern A;
*/
static void ecriretype(char *s, struct Type *info, int n, char *nom)
{
     int trouv;
     struct Type *p;
     
     p = info;
     for (trouv = 0 ; --n; info++)
        if (info->status != STOCK && info->status != ACCSMOD 
	                          && info->ntrouve) {
	   mstrcat(&s, info->nomtype);
           if (info->status == STRUCT || info->status == ENUM)
	      vastrcat(2, &s, " ", nom);
	   mstrcat(&s, " ");
	   trouv++;
	}
     if (!trouv)
        mstrcat(&s, "int ");
}

static int Tcomp(char **s, struct Type *t) { return strcmp(*s, t->nomtype);}
struct Type *dichoType(char *mot, const struct Type *tab, int n)
{
     return (struct Type *)bsearch((void *) &mot, (void *)tab, 
				   n, sizeof(*tab), (ARG) Tcomp);
}


inline static void add(struct Type *p, int *s, int n)
{
    while (--n >= 0)
        p++->ntrouve += *s++;
}

static *dupType(struct Type *t, int nt)
{
    int *pt;
    int *ini;
    pt = (int *) malloc(nt * sizeof(int));
    for (ini = pt; --nt >= 0;)
       *pt++ = t++->ntrouve;
    return ini;
}


#define TAILLEHACH 101
static struct Struc *tabhach[MACRO_INF+1][TAILLEHACH];

/* hacher : forme la valeur de hachage de la chaine s */
unsigned hacher(const char *s, int lim)
{
    unsigned valhach;

    for (valhach = 0; *s ; s++)
       valhach = *s + 31 * valhach;
    return valhach % lim;
}

/* consulter : try to find s in tabhach */
/* does not use the method of trouverDansArbre 
   as instastller push new variables in the begining
   of the stack.
*/

Struc *consulter(char *s, Entry level)
{
    struct Struc *pn;

    for (pn = tabhach[level][hacher(s,TAILLEHACH)]; pn != NULL; pn = pn->suiv)
       if (strcmp(s, pn->str.nom) == 0 && pn->str.status >= 0)
	     return pn;
    return NULL;        /* pas trouve */
}

/* installer : place (nom) dans tabhach */
static Struc *installer(char *nom, char *fnc, int i, Entry level)
{
    struct Struc *pn;
    unsigned valhach;

    if ((pn = consulter(nom, level)) == NULL || pn != NULL && 
	                  pn->str.status != nblocs || level == MACRO_INF) {
       pn = (struct Struc *) malloc(sizeof(*pn));
       if (pn == NULL || (pn->str.nom = dupchaine(nom)) == NULL ||
	   (pn->str.fonction = dupchaine(fnc)) == NULL)
	  AllocError();
       valhach = hacher(nom, TAILLEHACH);
       pn->suiv = tabhach[level][valhach];
       tabhach[level][valhach] = pn;
       if ((pn->str.nom = dupchaine(nom)) == NULL)
	  AllocError();
       pn->str.status = i;
       pn->str.ligne = nlignes;
       pn->str.file = filename;
           /* The default value of str.nligne */
       return pn;
    } else if (!pn->str.def) {
       pn->str.def = 1;
       return pn;
    } else if (pn != NULL)
       error("multiple defintion of %s, line %d", pn->str.nom, nlignes);
    return NULL;
}

       
Struc *defineType(struct Sortie *s, char *bloc,
		struct Type *type, int nelem, int status)
{
    Struc *pn;
    if ((pn = installer(s->var.nom, bloc, status, TP_INF)) == NULL ||
	(pn->obj.type = dupType(type, nelem)) == NULL)
       return NULL;
    pn->str.type = s->var.nom;
    pn->str.text = s->var.text;
    pn->str.str = s->var.str;
    pn->str.strtype = s->var.strtype;
    pn->str.def = 1;
    /* This is to avoid the confusion betwen void and void *. */
    if (pn->obj.type[Void])
       pn->obj.type[Void] = INDULGENT;
    return pn;
}

Struc *DefStruc(char *nom, int i, int def, char *func, Entry where)
{
    Struc *pn = installer(nom, func, i, where);
    if (pn == NULL)
       return NULL;
    pn->str.def = def;
    if (where == STR_INF) {
       pn->str.dec = NULL;
       pn->str.whatIs = STRUCT;
    }
    return pn;
}

/* set to OLD the status of all the elements in tabhach wich 
   are not in the current bloc */
void clean(void)
{
    int i;
    int j;
    struct Struc *pn;

    /* Macros do not have the same scope as structures/unions/typedefs ...
       hence the use of ENUM_INF instead of MACRO_INF */
    for (j = 0; j < ENUM_INF ; j++)
       for (i = 0; i < TAILLEHACH; i++) {
	  for (pn = tabhach[j][i]; pn != NULL; pn = pn->suiv) {
	     if (pn->str.status > nblocs)
	        pn->str.status = OLD;
	     /* to allow the use of structures defined in another file
		for example */
	  }
       }
}


void affDefStruc(const Condition *cond)
{
    int i;
    struct Struc *pn;

    printf("\n\n"
	   "\t/********************************************************\n"
	   "\t *        Definitions of structures and unions.         *\n"
           "\t ********************************************************/\n"
	   "\n\n");
    for (i = 0; i < TAILLEHACH; i++)
        for (pn = tabhach[STR_INF][i]; pn != NULL; pn = pn->suiv) {
	    if (cond->Affstruct != NULL && StrictTrouver(cond->Affstruct,
				  pn->str.nom, pn->str.fonction) == NULL)
	       continue;
	    putchar('\n');
	    printf(pn->str.whatIs == UNION ? " union" : " structure");
	    printf(" %s defined in %s (line %d of %s)\n",
		   pn->str.nom, pn->str.fonction, pn->obj.dec->var.ligne,
		   pn->obj.dec->var.file);
	    printf("which contains the following members :\n");
	    affDefMembers(pn->obj.dec);
	    printf(" end of the definition of %s in %s\n\n", 
		   pn->str.nom, pn->str.file);
	}
    putchar('\n');
}


void dispType(const Condition *cond)
{
    int i;
    Struc *pn;


    printf("\n\n"
	   "\t/********************************************************\n"
	   "\t *    Definitions of user defined types (typedefs).     *\n"
           "\t ********************************************************/\n"
	   "\n");

    for (i = 0; i < TAILLEHACH; i++) {
        for (pn = tabhach[TP_INF][i]; pn != NULL; pn = pn->suiv) {
	    if (cond->Afftypedef != NULL && StrictTrouver(cond->Afftypedef, 
			       pn->str.nom, pn->str.fonction) == NULL)
	       continue;
	    putchar('\n');
	    printf("%s is a type declared as : %s %s\n", pn->str.nom, 
		   pn->str.text == NULL ? ""/* nothing */ : pn->str.text,
		   pn->str.strtype);
	    printf("     defined in %s, line %d of %s\n", 
		   pn->str.fonction, pn->str.ligne, pn->str.file);
	}
    }
    putchar('\n');
}


void affDefEnum(const Condition *cond)
{
    int i;
    Struc *pn;

    printf("\n\n"
	   "\t/********************************************************\n"
	   "\t *           Definitions of enumerations (enum).        *\n"
           "\t ********************************************************/\n"
	   "\n");

    for (i = 0; i < TAILLEHACH; i++) {
        for (pn = tabhach[ENUM_INF][i]; pn != NULL; pn = pn->suiv) {
	    if (cond->Affenum != NULL && StrictTrouver(cond->Affenum,
			      pn->str.nom, pn->str.fonction) == NULL)
	       continue;
	    putchar('\n');
	    printf("%s is an enumeration that sets the following "
		   "constants :\n", pn->str.nom);
	    maketable(1, 5, 0, 65, pn->str.text);
	    printf("defined in %s (line %d of %s)\n",
		   pn->str.fonction, pn->str.ligne, pn->str.file);
	}
    }
    putchar('\n');
}


/*
   The following contains a miniature version of the C preprocessor.
*/

#define DEPTH_ARGS 7

/* treats the arguments  of a macro. */
static Args *consulterArgs(Args **arg, int lim, char *s)
{
     unsigned i = hacher(s, lim);
     Args *list = arg[i];

     for ( ; list != NULL; list = list->next)
        if (!strcmp(s, list->name))
	   return list;
     return NULL;
}

static Args *defineArgs(Args **arg, int lim, char *name, int count)
{
    int i;
    Args *cur;

    for (i = 0; i < lim; i++)
       for (cur = arg[i] ; cur != NULL ; cur = cur->next ) {
	  if (cur->n == count) {
	     cur->replace = dupchaine(name);
	     return cur;
	  }
       }
    return NULL;
}

static Args *installArgs(Args **arg, int lim, char *name, int count)
{
    int hach;
    Args *cur;

    if (consulterArgs(arg, lim, name) == NULL) {
       hach = hacher(name, lim);
       if ((cur = (Args *) malloc(sizeof(Args))) == NULL)
	  AllocError();
       cur->n = count;
       cur->next = arg[hach];
       cur->name = dupchaine(name);
       arg[hach] = cur;
       return arg[hach];
    } else
       fprintf(stderr, " the macro named %s line %d has two parameters with"
	       " the same name\n", name , nlignes);
    return NULL;
}


void affDefMacro(const Condition *cond)
{
    int i;
    Struc *pn;
    Args *res;

    printf("\n\n"
	   "\t/********************************************************\n"
	   "\t *          Definitions of macros ( #define ).          *\n"
           "\t ********************************************************/\n"
	   "\n");

    for (i = 0; i < TAILLEHACH; i++) {
        for (pn = tabhach[MACRO_INF][i]; pn != NULL; pn = pn->suiv) {
	    if (cond->Affmacro != NULL && StrictTrouver(cond->Affmacro,
			      pn->str.nom, pn->str.fonction) == NULL)
	       continue;
	    putchar('\n');
	    printf(" macro %s", pn->str.nom);
	    if (pn->obj.args != NULL) {
	       int j;
	       printf(" that takes as arguments :");
	       for (j = 1; (res = defineArgs(pn->obj.args, DEPTH_ARGS, "", j))
		            != NULL ; j++) {
		  printf("%s", (j == 1) ? "" : ",");
		  printf("%s", res->name);
	       }
	    }
	    printf(" :\n");
	    maketable(1, 5, 0, 65, pn->str.text);
	    printf(" used %d times, defined in %s, line %d of %s\n", 
	     pn->str.whatIs+1, pn->str.fonction, pn->str.ligne, pn->str.file);
	}
    }
    putchar('\n');
}

void affDefVar(Variable *p, const Condition *cond)
{
     if (p != NULL) {
        printf("\n\n"
	   "\t/********************************************************\n"
	   "\t *               Definitions of Variables.              *\n"
           "\t ********************************************************/\n"
	   "\n");
	affarbre(p, cond);
     }
}


/*
    Functions to treat the #define and the #undef directive.
*/

static int Define(char *);
static int Undef(char *);

struct Macro {
    char *name;
    int (*operations)(char *);
} static macro[] = 
{
   "#define", Define,
   "#undef", Undef
};

static int const NKEYS = sizeof(macro) / sizeof(macro[0]);

typedef Args *(*ToFind)(Args **, int depth, char *name, int count);

static int readArgs(Struc *Macro, char **name, int *size, ToFind find)
{
    int count;
    int c;
    char *arg = NULL;

    for (count = 1; ; count++, arg = NULL) {
          while ((c = readword(name, size)) != ',' && c != ')' && c != EOF) {
	     if (arg != NULL)
	        mstrcat(&arg, " ");
	     mstrcat(&arg, *name);
	  }
	  if (!(*find)(Macro->obj.args, DEPTH_ARGS, arg, count))
	     return 0;
	  if (c != ',' && c != ')') {
	     remettrecar(0); /* cause a syntax error */
	     free((void *) arg);
	     return 0;
	  } else if (c == ')')
	     break;
	  free((void *) arg);
    }
    free((void *) arg);
    return 1;
}

static int Define(char *func)
{
    char *name = NULL;
    char newc[2];
    int size = 0, ligne;
    int defSize = 0, textSize = 0, c;
    const step = 100;
    Struc *Macro;

    newc[1] = '\0';
    if (!isalpha(c = readword(&name, &size)) && c != '_') {
       fprintf(stderr, 
	       " incorrect variable name : %s line %d\n", name, nlignes);
       return 0;
    }
    if ((Macro = installer(name, func, 0, MACRO_INF)) == NULL)
       return 0;
    ligne = nlignes;
    /* Macro with arguments */
    if ((c = lirecar()) == '(') {
       int count;
       if ((Macro->obj.args = (Args **) malloc((sizeof(Args *) * DEPTH_ARGS))) 
	              == NULL)
	  AllocError();
       for (count = 0; count < DEPTH_ARGS ; count++)
	  Macro->obj.args[count] = NULL;
       if (!readArgs(Macro, &name, &size, installArgs))
	  return 0;
    } else {
       Macro->obj.args = NULL;
       remettrecar(c);
    }
    Macro->str.dec = Macro->str.text = 0;
    /* if ther number of lines is the same in all the function */
    if (ligne == nlignes) {
       for (textSize = defSize = step ; (c = lirecar()) != '\n' ;) {
	  for (;;) {
	     /* does not consider comments, quotes  */
	    int res1, res2;
	    res1 = negligerComment(c);
	    res2 = negligerGuillemets(c);
	    if (res1 || res2) {
	       if (res2)
		  mistrcat(&Macro->str.text, getQuote(), &textSize);
	       if ((c = lirecar()) == '\n')
		  break;
	    } else
	       break;
	  }
	  if (c == '\n')
	     break;
	  while (c == '\\') {
	     while ((c = lirecar()) != '\n' && c != EOF)
	        /* wait */;
	     if (c == EOF)
	        break;
	     nlignes++;
	     c = lirecar();
	  }
	  if (c == EOF) {
	     remettrecar(0); /* cause a syntax error */
	     return 0;
	  }
	  *newc = c;
	  mistrcat(&Macro->str.dec, newc, &defSize);
	  mistrcat(&Macro->str.text, newc, &textSize);
       }
       nlignes++;
    } else 
       Macro->str.dec = NULL;
    Macro->str.ligne = nlignes -1; /* for a macro the newline is the end */
    Macro->str.def = 0;            /* of the definition. */
    Macro->str.whatIs = 0;
    return 1;
}

static int Undef(char *func)
{
    Struc *Macro;
    char *s = NULL;
    int sSize = 0;
    int c;

    if (!isalpha(c = readword(&s, &sSize)) && c != '_' && c != SPECIAL) {
        fprintf(stderr, " wrong macro name %s bloc %s line %d\n", 
		          s, func, nlignes);
	return 0;
    }
    if ((Macro = consulter(s, MACRO_INF)) != NULL)
       Macro->str.status = OLD;
    else
       fprintf(stderr, " trying to delete the value of the undefined macro %s"
	       " line %d\n", s, nlignes);
    return 1;
}

static int cmp(char *s, struct Macro *p) { return strcmp(s, p->name); }
int findMacro(char *nm, char *func)
{
    struct Macro *i;

    if (i = (struct Macro *) bsearch((void *) nm, (void *) macro, 
         NKEYS, sizeof(macro[0]), (ARG) cmp))
	 return (i->operations)(func);
    return 1;
}

int OpenMacro(char *name)
{
    Struc *macro;
    Args *arg;
    int size = 0, c, curlig;
    char *args = NULL, *replace = NULL;

    if ((macro = consulter(name, MACRO_INF)) != NULL) {
       macro->str.def = nlignes;
       macro->str.whatIs++;
       if (macro->obj.args != NULL) {
	  if (readword(&args , &size) != '(') {
	     fprintf(stderr, " argument list expected line %d macro %s\n",
		             nlignes, macro->str.nom);
	     remettrech(args);
	     return 0;
	  }
	  if (!readArgs(macro, &args, &size, defineArgs)) {
	     fprintf(stderr, " incorrect parameters to macro %s line %d\n",
		     macro->str.nom , nlignes);
	     return 0;
	  }
       }
       remettrecar(0);  /* to know when the preprocessor text is finished */
       /* store the current line number */
       curlig = nlignes;
       remettrech(macro->str.dec);
       while (c = readword(&args, &size)) {
	  if (macro->obj.args != NULL)
	     if (arg = consulterArgs(macro->obj.args, DEPTH_ARGS, args)) {
	        vastrcat(2, &replace, arg->replace, " ");
		continue;
	     }
	  vastrcat(2, &replace, args, " ");
       }
       /* restore the number of lines */
       nlignes = curlig;
       if (replace != NULL)
	  remettrech(replace);
       return 1;
    }
    return 0;
}


#undef DEPTH_ARGS
#undef TAILLEHACH

